
from docido_sdk.core import Interface

__all__ = [
    'ICrawler',
]


class ICrawler(Interface):
    """Extension point interface for components willing to
    provide additional Docido crawlers.
    """
    def get_service_name():
        """
        :return: crawler name
        :rtype: string
        """

    def iter_crawl_tasks(index, oauth_token, logger, full):
        """Split the crawl in smaller independant actions,
        and returns them for delayed execution.

        :param docido_sdk.push.api.IndexAPI: index
          To manipulate Docido index

        :param docido_sdk.oauth.OAuthToken oauth_token:
          OAuth credentials

        :param logging.Logger logger:
          to emit messages

        :param bool full:
          whether the entire account must be pushed or only
          changes that occured since previous crawl.

        :return: generator of :py:func:`functools.partial` tasks
          to execute to perform the account synchronization.
          partial objects may accept 2 arguments:

          - push_api (:py:class:`docido_sdk.push.IndexAPI`)
          - oauth_token (:py:class:`docido_sdk.oauth.OAuthToken`)
          - logger (:py:class:`logging.Logger`)

          A tuple of 2 elements can also be returned for crawlers
          willing to perform a final operation when all sub-tasks
          have been executed. The tuple may be like:
          `tuple(generator of partial, partial)`

          A task cannot be instance, class, or static method
          of a :py:func:`docido_sdk.core.Component` object.
          Therefore you may provide functions defined outside
          your crawler class definition.
        """

    def clear_account(index, oauth_token, logger):
        """Remove from Docido index all data previously indexed for
        this account. Persisted data must also be cleared.

        :param docido_sdk.push.IndexAPI index:
          To manipulate Docido index

        :param docido_sdk.oauth.OAuthToken oauth_token:
          OAuth credentials

        :param logging.Logger logger:
          to emit messages
        """




from contextlib import contextmanager
import copy


@contextmanager
def restore_dict_kv(a_dict, key, copy_func=copy.deepcopy):
    """Backup an object in a with context and restore it when leaving
    the scope.

    :param a_dict:
      associative table
    :param: key
      key whose value has to be backed up
    :param copy_func: callbable object used to create an object copy.
    default is `copy.deepcopy`
    """
    backup = copy_func(a_dict[key])
    try:
        yield
    finally:
        a_dict[key] = backup



import errno
import os
import os.path as osp
import sys

import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


from . toolbox.collections_ext import nameddict, contextobj


class Configuration(nameddict):
    @classmethod
    def from_file(cls, path):
        if not osp.exists(path) and not osp.isabs(path):
            path = osp.join(osp.dirname(osp.abspath(__file__)), path)
        with open(path, 'r') as istr:
            return Configuration(yaml.load(istr, Loader=Loader))

try:
    config = Configuration.from_file(
        os.getenv('DOCIDO_CONFIG', 'settings.yml')
    )
except IOError as e:
    # Issue happened while loading settings.yml
    if e.errno == errno.ENOENT:
        # Ignore issue only if file was not found
        config = Configuration()
    else:
        raise e

sys.modules[__name__] = contextobj(config)
# This module is now the configuration object itself.



"""
Provides a set of Exception classes a crawler may raise when it is
not possible to consume source API because of OAuth-related issues.
"""


class CrawlerError(Exception):
    """Common base exception class for crawl issues"""
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          reason for error.
        """
        super(CrawlerError, self).__init__(message)


class OAuthTokenExpiredError(CrawlerError):
    """Exception class used to notify crawler manager when
    the OAuth token of an account has expired, and needs to be refreshed
    thru Docido front application.
    """
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          the error message returned by source API.
        """
        super(OAuthTokenExpiredError, self).__init__(message)


class OAuthTokenPermanentError(CrawlerError):
    """Exception class used to notify crawler manager when
    the OAuth token of an account has been revoked.
    """
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          the error message returned by source API.
        """
        super(OAuthTokenPermanentError, self).__init__(message)


class OAuthTokenRefreshRequiredError(CrawlerError):
    """Exception class used to notify crawler manager when
    the OAuth token of an account has expired or will expire
    shortly, and need to be refreshed programmaticaly.
    """
    def __init__(self, message=None):
        """
        @param message: optional printable object providing
          the error message returned by source API.
        """
        super(OAuthTokenRefreshRequiredError, self).__init__(message)



import logging
import os
import os.path as osp
import pickle
from pickle import PickleError

import yaml

from .. import loader
from ..env import env
from ..oauth import OAuthToken
from ..core import (
    implements,
    Component,
    ExtensionPoint,
)
from ..crawler import ICrawler
from ..index.config import YamlPullCrawlersIndexingConfig
from ..index.test import LocalKV, LocalDumbIndex
from ..index.processor import (
    Elasticsearch,
    CheckProcessor,
)
from ..index.pipeline import (
    IndexPipelineProvider,
    IndexAPIConfigurationProvider,
)

import docido_sdk.config as docido_config
from ..toolbox.collections_ext import nameddict


class YamlAPIConfigurationProvider(Component):
    implements(IndexAPIConfigurationProvider)

    def get_index_api_conf(self, service, docido_user_id, account_login):
        return {
            'service': service,
            'docido_user_id': docido_user_id,
            'account_login': account_login
        }


def oauth_tokens_from_file(full=True, config=None):
    oauth_path = os.getenv('DOCIDO_CC_RUNS', '.dcc-runs.yml')
    with open(oauth_path, 'r') as oauth_file:
        crawlers = nameddict(yaml.load(oauth_file))
        for crawler, runs in crawlers.iteritems():
            for run, run_config in runs.iteritems():
                run_config.token = OAuthToken(**run_config.token)
                run_config.setdefault('config', config)
                run_config.setdefault('full', full)
        return crawlers


def load_yaml(path):
    if not osp.exists(path) and not osp.isabs(path):
        path = osp.join(osp.dirname(osp.abspath(__file__)), path)
    with open(path, 'r') as istr:
        return nameddict(yaml.load(istr))


class LocalRunner(Component):
    crawlers = ExtensionPoint(ICrawler)

    def _check_pickle(self, tasks):
        return pickle.dumps(tasks)

    def run(self, full=False, config=None):
        crawler_runs = oauth_tokens_from_file(full=full, config=config)
        index_pipeline_provider = env[IndexPipelineProvider]
        for crawler, launches in crawler_runs.iteritems():
            c = [c for c in self.crawlers if c.get_service_name() == crawler]
            if len(c) != 1:
                raise Exception(
                    'unknown crawler for service: {}'.format(crawler)
                )
            c = c[0]
            for launch, launch_config in launches.iteritems():
                logger = logging.getLogger(
                    '{crawler}.{launch}'.format(crawler=crawler, launch=launch)
                )
                with docido_config._push():
                    if launch_config.config is not None:
                        docido_config.clear()
                        docido_config.update(load_yaml(launch_config.config))

                    index_api = index_pipeline_provider.get_index_api(
                        crawler, None, None
                    )
                    tasks = c.iter_crawl_tasks(index_api, launch_config.token,
                                               logger, launch_config.full)
                    try:
                        self._check_pickle(tasks)
                    except PickleError as e:
                        raise Exception(
                            'unable to serialize crawl tasks: {}'.format(
                                str(e)
                            )
                        )

                    def _runtask(task):
                        task(index_api, launch_config.token, logger)

                    def _runtasks(tasks):
                        for t in tasks:
                            t(index_api, launch_config.token, logger)

                    _runtasks(tasks['tasks'])
                    if 'epilogue' in tasks:
                        _runtask(tasks['epilogue'])


def run(*args):
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option(
        '-i',
        action='store_true',
        dest='incremental',
        help='trigger incremental crawl'
    )
    parser.add_option(
        '-v', '--verbose',
        action='count',
        dest='verbose',
        help='set verbosity level'
    )
    (options, args) = parser.parse_args()
    verbose = options.verbose if options.verbose is not None else 0
    logging_level = logging.WARN
    if verbose == 1:
        logging_level = logging.INFO
    elif verbose > 1:
        logging.level = logging.DEBUG
    logging.basicConfig(level=logging_level)

    loader.load_components(env)
    env[YamlPullCrawlersIndexingConfig]
    env[Elasticsearch]
    env[CheckProcessor]
    env[IndexPipelineProvider]
    env[LocalKV]
    env[LocalDumbIndex]
    runner = env[LocalRunner]
    runner.run(full=not options.incremental)
